/**
 * @file    shellmatta_ymodem.h
 * @brief   ymodem functions of shellmatta
 * @author  Simon Fischer <dev@s-fischer.net>
 */

#ifndef _SHELLMATTA_YMODEM_H_
#define _SHELLMATTA_YMODEM_H_

#include <stdint.h>
#include "shellmatta.h"

#define YMODEM_PACKET_SIZE          128u    /* default packet size of ymodem transmission */
#define YMODEM_PACKET_SIZE_1K       1024u   /* extended packet size of ymodem transmission */
#define YMODEM_HEADER_DATA_OFFSET   2u      /* offset position of data field in header packet */
#define YMODEM_HEADER_CRC_POSITION  130u    /* position of crc16 in header packet, relative to packet number */

/** @brief symbols needed for ymodem protocol */
typedef enum {
    YMODEM_NULL    = 0x00, /* NULL-terminator */
    YMODEM_SOH     = 0x01, /* Start of header */
    YMODEM_STX     = 0x02, /* Start of header for 1k packet */
    YMODEM_EOT     = 0x04, /* End of transmission */
    YMODEM_ACK     = 0x06, /* Acknowledge */
    YMODEM_NAK     = 0x15, /* Negative acknowledge */
    YMODEM_CA      = 0x18, /* Cancel */
    YMODEM_SPACE   = 0x20, /* Space */
    YMODEM_CRC     = 0x43  /* 'C' to indicate CRC type */
} YMODEM_SYMBOLS;

/** @brief packet structure that holds several information about its content */
typedef struct
{
    uint8_t packetNumber;           /* packet number in this packet */
    uint8_t reversePacketNumber;    /* reverse packet number in this packet */
    uint16_t packetCrc;             /* crc checksum in this packet */
    uint8_t* packetData;            /* pointer to the data of this packet */
    uint16_t* packetSize;           /* pointer to the length of this packet */
    uint32_t* fileSize;             /* pointer to the amount of bytes in this packet */
} shellmatta_ymodem_packet_t;

/**
 * @brief return codes for receive byte function of ymodem module
*/
typedef enum {
    SOH_RECEIVED,
    STX_RECEIVED,
    EOT_RECEIVED,
    CA_RECEIVED,
    DATA_RECEIVED,
    ERROR
} shellmatta_ymodem_rcv_retcode_t;

typedef enum {
    YMODEM_NONE,   /* no data is transmitted */
    YMODEM_HEADER, /* header packet is transmittet */
    YMODEM_BODY,   /* data packet is transmitted */
    YMODEM_FOOTER  /* end of transmission packet is transmitted */
} shellmatta_ymodem_datatype_t;

void shellmatta_ymodem_init(            shellmatta_handle_t             handle, 
                                        uint8_t*                        recvBuffer, 
                                        uint32_t*                       fileSize,
                                        uint16_t*                       packetSize,
                                        shellmatta_ymodem_callbacks_t   callbacks);

void shellmatta_ymodem_receive_packet(  shellmatta_handle_t             handle, 
                                        uint8_t                         byteIn);

void shellmatta_ymodem_set_state(       shellmatta_handle_t             handle, 
                                        shellmatta_ymodem_state_t       newState);

shellmatta_ymodem_state_t shellmatta_ymodem_get_state(          shellmatta_handle_t handle);

shellmatta_ymodem_rcv_retcode_t shellmatta_ymodem_receive_byte( shellmatta_handle_t handle, 
                                                                uint8_t             byteIn);

void shellmatta_ymodem_control(         shellmatta_handle_t handle, 
                                        const char          c);

void shellmatta_ymodem_ack(             shellmatta_handle_t handle);

void shellmatta_ymodem_nak(             shellmatta_handle_t handle);

void shellmatta_ymodem_check_packet(    shellmatta_handle_t handle);

void shellmatta_ymodem_reset(           shellmatta_handle_t handle,
                                        bool                doCancel);

shellmatta_ymodem_datatype_t shellmatta_ymodem_get_current_datatype(void);

void shellmatta_ymodem_enable(void);

void shellmatta_ymodem_disable(void);

#endif /* _SHELLMATTA_YMODEM_H_ */
